"""Flexible data structures, ideal to specify nested project configurations."""

from .jsonobject import JsonObject
from .jsonobjects import JsonObjects

__all__ = [
    'JsonObject',
    'JsonObjects'
]
